@@ -26,7 +26,7 @@ no warnings qw(recursion qw);
 use strict;
 
 use vars qw($VERSION $STRING_VERSION);
-$VERSION        = '2.067_002';
+$VERSION        = '2.067_003';
 $STRING_VERSION = $VERSION;
 ## no critic(BuiltinFunctions::ProhibitStringyEval)
 $VERSION = eval $VERSION;
@@ -648,6 +648,30 @@ sub Marpa::R2::Grammar::show_problems {
     return "Grammar has no problems\n";
 } ## end sub Marpa::R2::Grammar::show_problems
 
+# Return DSL name of symbol
+# Does no checking
+sub Marpa::R2::Grammar::symbol_dsl_name {
+    my ( $grammar, $symbol_id ) = @_;
+    my $symbols   = $grammar->[Marpa::R2::Internal::Grammar::SYMBOLS];
+    my $symbol = $symbols->[$symbol_id];
+    return [Marpa::R2::Internal::Symbol::DSL_NAME];
+}
+
+# Return display form of symbol
+# Does lots of checking and makes use of alternatives.
+sub Marpa::R2::Grammar::symbol_in_display_form {
+    my ( $grammar, $symbol_id ) = @_;
+    my $symbols   = $grammar->[Marpa::R2::Internal::Grammar::SYMBOLS];
+    my $symbol = $symbols->[$symbol_id];
+    return "<!No symbol with ID $symbol_id!>" if not defined $symbol;
+    my $text = $symbol->[Marpa::R2::Internal::Symbol::DISPLAY_FORM];
+    return $text if defined $text;
+    $text = $symbol->[Marpa::R2::Internal::Symbol::DSL_NAME];
+    return "<$text>" if defined $text;
+    $text = $grammar->symbol_name($symbol_id);
+    return "<$text>";
+}
+
 sub Marpa::R2::Grammar::show_symbol {
     my ( $grammar, $symbol ) = @_;
     my $grammar_c = $grammar->[Marpa::R2::Internal::Grammar::C];
@@ -736,16 +760,6 @@ sub Marpa::R2::Grammar::brief_rule {
     return ( join q{ }, "$rule_id:", $lhs, '->', @rhs ) . $quantifier;
 } ## end sub Marpa::R2::Grammar::brief_rule
 
-sub Marpa::R2::Grammar::brief_original_rule {
-    my ( $grammar, $rule_id ) = @_;
-    my $grammar_c = $grammar->[Marpa::R2::Internal::Grammar::C];
-    my $tracer = $grammar->[Marpa::R2::Internal::Grammar::TRACER];
-    my ( $lhs, @rhs ) = $tracer->rule($rule_id);
-    my $minimum = $grammar_c->sequence_min($rule_id);
-    my $quantifier = defined $minimum ? $minimum <= 0 ? q{*} : q{+} : q{};
-    return ( join q{ }, "$rule_id:", $lhs, '->', @rhs ) . $quantifier;
-} ## end sub Marpa::R2::Grammar::brief_original_rule
-
 sub Marpa::R2::Grammar::show_rule {
     my ( $grammar, $rule ) = @_;
 
@@ -793,6 +807,7 @@ sub Marpa::R2::Grammar::rule_ids {
 sub Marpa::R2::Grammar::rule {
     my ( $grammar, $rule_id ) = @_;
     my $grammar_c   = $grammar->[Marpa::R2::Internal::Grammar::C];
+    my $symbols     = $grammar->[Marpa::R2::Internal::Grammar::SYMBOLS];
     my $rule_length = $grammar_c->rule_length($rule_id);
     return if not defined $rule_length;
     my @symbol_ids = ( $grammar_c->rule_lhs($rule_id) );
@@ -800,12 +815,13 @@ sub Marpa::R2::Grammar::rule {
         map { $grammar_c->rule_rhs( $rule_id, $_ ) }
         ( 0 .. $rule_length - 1 );
     my @symbol_names = ();
-    for my $symbol_id (@symbol_ids) {
+    SYMBOL_ID: for my $symbol_id (@symbol_ids) {
         ## The name of the symbols, before the BNF rewrites
-        push @symbol_names,
-            Marpa::R2::Grammar::original_symbol_name(
-            $grammar->symbol_name($symbol_id) );
-    } ## end for my $symbol_id (@symbol_ids)
+        my $name =
+            $symbols->[$symbol_id]->[Marpa::R2::Internal::Symbol::LEGACY_NAME]
+            // $grammar->symbol_name($symbol_id);
+        push @symbol_names, $name;
+    } ## end SYMBOL_ID: for my $symbol_id (@symbol_ids)
     return @symbol_names;
 } ## end sub Marpa::R2::Grammar::rule
 
@@ -916,6 +932,26 @@ sub assign_symbol {
             $grammar_c->symbol_rank_set($symbol_id) = $value;
             next PROPERTY;
         } ## end if ( $property eq 'rank' )
+        if ( $property eq 'description' ) {
+            my $value = $options->{$property};
+            $symbol->[Marpa::R2::Internal::Symbol::DESCRIPTION] = $value;
+            next PROPERTY;
+        }
+        if ( $property eq 'dsl_name' ) {
+            my $value = $options->{$property};
+            $symbol->[Marpa::R2::Internal::Symbol::DSL_NAME] = $value;
+            next PROPERTY;
+        }
+        if ( $property eq 'legacy_name' ) {
+            my $value = $options->{$property};
+            $symbol->[Marpa::R2::Internal::Symbol::LEGACY_NAME] = $value;
+            next PROPERTY;
+        }
+        if ( $property eq 'display_form' ) {
+            my $value = $options->{$property};
+            $symbol->[Marpa::R2::Internal::Symbol::DISPLAY_FORM] = $value;
+            next PROPERTY;
+        }
         Marpa::R2::exception(qq{Unknown symbol property "$property"});
     } ## end PROPERTY: for my $property ( keys %{$options} )
 
@@ -933,12 +969,14 @@ sub assign_user_symbol {
         Marpa::R2::exception(
             "Symbol name was ref to $type; it must be a scalar string");
     }
-    my $final_symbol = substr $name, -1;
-    if ( $DEFAULT_SYMBOLS_RESERVED{$final_symbol} ) {
-        Marpa::R2::exception(
-            qq{Symbol name $name ends in "$final_symbol": that's not allowed}
-        );
-    }
+    if ( not $grammar->[Marpa::R2::Internal::Grammar::INTERNAL] ) {
+        my $final_symbol = substr $name, -1;
+        if ( $DEFAULT_SYMBOLS_RESERVED{$final_symbol} ) {
+            Marpa::R2::exception(
+                qq{Symbol name $name ends in "$final_symbol": that's not allowed}
+            );
+        }
+    } ## end if ( not $grammar->[Marpa::R2::Internal::Grammar::INTERNAL...])
     my $symbol = assign_symbol( $grammar, $name, $options );
 
     return $symbol;
@@ -1016,6 +1054,7 @@ sub add_user_rule {
     my $mask;
     my $proper_separation = 0;
     my $keep_separation   = 0;
+    my $description;
 
     OPTION: while ( my ( $option, $value ) = each %{$options} ) {
         if ( $option eq 'name' )   { $rule_name = $value; next OPTION; }
@@ -1039,6 +1078,7 @@ sub add_user_rule {
         }
         if ( $option eq 'keep' ) { $keep_separation = $value; next OPTION }
         if ( $option eq 'mask' ) { $mask = $value; next OPTION }
+        if ( $option eq 'description' ) { $description = $value; next OPTION }
         Marpa::R2::exception("Unknown user rule option: $option");
     } ## end OPTION: while ( my ( $option, $value ) = each %{$options} )
 
@@ -1184,6 +1224,10 @@ sub add_user_rule {
             $ordinary_rule->[Marpa::R2::Internal::Rule::BLESSING] = $blessing;
         }
 
+        if ( defined $description ) {
+            $ordinary_rule->[Marpa::R2::Internal::Rule::DESCRIPTION] = $description;
+        }
+
         return;
     }    # not defined $min
 
@@ -1241,8 +1285,11 @@ sub add_user_rule {
         $rules_by_name->{$rule_name} = $original_rule;
     }
     if ( defined $blessing ) {
-            $original_rule->[Marpa::R2::Internal::Rule::BLESSING] = $blessing;
-            }
+        $original_rule->[Marpa::R2::Internal::Rule::BLESSING] = $blessing;
+    }
+    if ( defined $description ) {
+        $original_rule->[Marpa::R2::Internal::Rule::DESCRIPTION] = $description;
+    }
 
     return;
 
